# UnsafeAssume

[![PkgEval](https://JuliaCI.github.io/NanosoldierReports/pkgeval_badges/U/UnsafeAssume.svg)](https://JuliaCI.github.io/NanosoldierReports/pkgeval_badges/U/UnsafeAssume.html)

Helps with convincing the Julia compiler (LLVM, to be specific) of the truth of some
condition, statement, predicate, observation etc. Using this is **dangerous**, only
use when certain that:

1. the forced assumption is correct, i.e., don't *lie to the compiler*

2. the performance benefit is significant for you

Due to the limitations of the current implementations of both Julia and LLVM, any
performance improvements achieved may be extremely fragile.

Implemented using raw LLVM assembly code with
[`Core.Intrinsics.llvmcall`](https://docs.julialang.org/en/v1/base/c/#Core.Intrinsics.llvmcall).

Similar functionality exists in C++ and Rust:

* https://en.cppreference.com/w/cpp/utility/unreachable

* https://en.cppreference.com/w/cpp/language/attributes/assume

* https://en.cppreference.com/w/cpp/memory/assume_aligned

* https://doc.rust-lang.org/std/hint/fn.unreachable_unchecked.html

## Usage examples

The calls of `UnsafeAssume` functions, along with all the code that establishes the
relevant assumptions, together with the code where the applying the assumptions is
necessary for performance, should be within an `@inline` block. This ensures all
relevant information is available to the LLVM optimizer when the optimizer needs it.

### Preventing the domain check for `sqrt`

If we convince LLVM that a real floating-point argument to `sqrt` is nonnegative,
it may eliminate the code for throwing the domain error, resulting in the compiled
code consisting only of the essential square root instruction (`vsqrtsd`, in this
case), apart from the stack management instructions that any function needs to
have:

```julia-repl
julia> using UnsafeAssume

julia> function unsafe_sqrt(x)
         @inline begin
           is_negative = x < zero(x)
           unsafe_assume_condition(!is_negative)
           sqrt(x)
         end
       end
unsafe_sqrt (generic function with 1 method)

julia> Base.print_statement_costs(stdout, unsafe_sqrt, Tuple{Float64})
unsafe_sqrt(x) @ Main REPL[2]:1
    0 1 ─ %1  = Base.lt_float(_2, 0.0)::Bool
    0 │   %2  = Base.not_int(%1)::Bool
    0 └──       goto #3 if not %2
    0 2 ─       goto #4
 1000 3 ─       (Core.Intrinsics.llvmcall)("unreachable", UnsafeAssume.Cvoid, Tuple{})::Nothing
    0 4 ┄       goto #5
    0 5 ─ %7  = Base.lt_float(_2, 0.0)::Bool
    0 └──       goto #7 if not %7
    0 6 ─       invoke Base.Math.throw_complex_domainerror(:sqrt::Symbol, _2::Float64)::Union{}
    0 └──       unreachable
   20 7 ─ %11 = Base.Math.sqrt_llvm(_2)::Float64
    0 └──       goto #8
    0 8 ─       return %11
      

julia> @code_llvm unsafe_sqrt(0.3)
; Function Signature: unsafe_sqrt(Float64)
;  @ REPL[2]:1 within `unsafe_sqrt`
define double @julia_unsafe_sqrt_6500(double %"x::Float64") #0 {
top:
;  @ REPL[2]:3 within `unsafe_sqrt`
; ┌ @ float.jl:587 within `<`
   %0 = fcmp uge double %"x::Float64", 0.000000e+00
; └
;  @ REPL[2]:4 within `unsafe_sqrt`
; ┌ @ /some/path/UnsafeAssume/src/UnsafeAssume.jl:36 within `unsafe_assume_condition`
   call void @llvm.assume(i1 %0)
; └
;  @ REPL[2]:5 within `unsafe_sqrt`
; ┌ @ math.jl:687 within `sqrt`
   %1 = call double @llvm.sqrt.f64(double %"x::Float64")
; └
  ret double %1
}

julia> @code_native debuginfo=:none unsafe_sqrt(0.3)
        .text
        .file   "unsafe_sqrt"
        .globl  julia_unsafe_sqrt_6726          # -- Begin function julia_unsafe_sqrt_6726
        .p2align        4, 0x90
        .type   julia_unsafe_sqrt_6726,@function
julia_unsafe_sqrt_6726:                 # @julia_unsafe_sqrt_6726
; Function Signature: unsafe_sqrt(Float64)
# %bb.0:                                # %top
        #DEBUG_VALUE: unsafe_sqrt:x <- $xmm0
        push    rbp
        mov     rbp, rsp
        vsqrtsd xmm0, xmm0, xmm0
        pop     rbp
        ret
```

### Preventing an undefined reference check

Arrays with elements of `mutable` type may check whether the element is an
undefined reference on each access. This rarely matter for performance, but here's
how to prevent it anyway:

```julia-repl
julia> function unsafe_array_access(a)
         @inline begin
           unsafe_assume_condition(isassigned(a, 1))
           @inbounds a[1]
         end
       end
unsafe_array_access (generic function with 1 method)

julia> a = [Int[]]  # This array type may contain undefined references because its elements are mutable
1-element Vector{Vector{Int64}}:
 []

julia> @code_llvm unsafe_array_access(a)
; Function Signature: unsafe_array_access(Array{Array{Int64, 1}, 1})
;  @ REPL[7]:1 within `unsafe_array_access`
define nonnull {}* @julia_unsafe_array_access_7773({}* noundef nonnull align 16 dereferenceable(40) %"a::Array") #0 {
top:
;  @ REPL[7]:3 within `unsafe_array_access`
; ┌ @ array.jl:268 within `isassigned`
; │┌ @ abstractarray.jl:684 within `checkbounds`
; ││┌ @ abstractarray.jl:386 within `eachindex`
; │││┌ @ abstractarray.jl:134 within `axes1`
; ││││┌ @ abstractarray.jl:98 within `axes`
; │││││┌ @ array.jl:191 within `size`
        %0 = bitcast {}* %"a::Array" to { i8*, i64, i16, i16, i32 }*
        %.length_ptr = getelementptr inbounds { i8*, i64, i16, i16, i32 }, { i8*, i64, i16, i16, i32 }* %0, i64 0, i32 1
        %.length = load i64, i64* %.length_ptr, align 8
; ││└└└└
; ││┌ @ abstractarray.jl:760 within `checkindex`
; │││┌ @ int.jl:513 within `<`
      %.not = icmp ne i64 %.length, 0
; │└└└
   call void @llvm.assume(i1 %.not)
; │ @ array.jl:270 within `isassigned`
   %1 = bitcast {}* %"a::Array" to {}***
   %.data5 = load {}**, {}*** %1, align 8
   %array_slot = load atomic {}*, {}** %.data5 unordered, align 8
; └
;  @ REPL[7]:4 within `unsafe_array_access`
  ret {}* %array_slot
}

julia> @code_native debuginfo=:none unsafe_array_access(a)
        .text
        .file   "unsafe_array_access"
        .globl  julia_unsafe_array_access_7783  # -- Begin function julia_unsafe_array_access_7783
        .p2align        4, 0x90
        .type   julia_unsafe_array_access_7783,@function
julia_unsafe_array_access_7783:         # @julia_unsafe_array_access_7783
; Function Signature: unsafe_array_access(Array{Array{Int64, 1}, 1})
# %bb.0:                                # %top
        #DEBUG_VALUE: unsafe_array_access:a <- [DW_OP_deref] $rdi
        #DEBUG_VALUE: unsafe_array_access:a <- [DW_OP_deref] 0
        push    rbp
        mov     rax, qword ptr [rdi]
        mov     rbp, rsp
        mov     rax, qword ptr [rax]
        pop     rbp
        ret
```
