# Copyright © 2023 Neven Sajko. All rights reserved.

module UnsafeAssume

export
  unsafe_assume_this_call_is_unreachable, unsafe_assume_condition,
  unsafe_assume_integer_is_divisible_by_power_of_two,
  unsafe_assume_pointer_is_aligned

"""
LLVM will assume that any call of this function will *never* be reached in any
execution of your code. In other words, this is assumed to be *unreachable*.

!!! danger
    Lying to the compiler is **dangerous**, convincing LLVM that a false statement
    is true can break your program in subtle ways.
"""
@inline function unsafe_assume_this_call_is_unreachable()
  @inline begin
    # https://llvm.org/docs/LangRef.html#unreachable-instruction
    Core.Intrinsics.llvmcall("unreachable", Cvoid, Tuple{})
    nothing
  end
end

"""
LLVM will assume that the provided argument will be `true`, in any possible
execution of your program.

!!! danger
    Lying to the compiler is **dangerous**, convincing LLVM that a false statement
    is true can break your program in subtle ways.
"""
@inline function unsafe_assume_condition(c::Bool)
  @inline begin
    # TODO: implement using LLVM's `llvm.assume` intrinsic directly?
    # https://llvm.org/docs/LangRef.html#llvm-assume-intrinsic
    c || unsafe_assume_this_call_is_unreachable()
    nothing
  end
end

function integer_is_divisible_by_power_of_two(n::Integer, ::Val{mask}) where {mask}
  @inline begin
    mask::Unsigned
    T = typeof(mask)
    rough = n % T
    finer = rough & mask
    iszero(finer)::Bool
  end
end

"""
LLVM will assume that the provided integer `n` has its least significant bits
cleared, as specified by the unsigned value `mask`. The latter is passed in the type
domain, using `Val`.

!!! danger
    Lying to the compiler is **dangerous**, convincing LLVM that a false statement
    is true can break your program in subtle ways.
"""
@inline function unsafe_assume_integer_is_divisible_by_power_of_two(n::Integer, mask::Val)
  @inline begin
    is_round = integer_is_divisible_by_power_of_two(n, mask)
    unsafe_assume_condition(is_round)
    nothing
  end
end

"""
LLVM will assume that the provided pointer `p` has its least significant bits
cleared, as specified by the unsigned value `mask`. The latter is passed in the type
domain, using `Val`.

!!! danger
    Lying to the compiler is **dangerous**, convincing LLVM that a false statement
    is true can break your program in subtle ways.
"""
@inline function unsafe_assume_pointer_is_aligned(p::Ptr, mask::Val)
  @inline begin
    n = reinterpret(UInt, p)
    unsafe_assume_integer_is_round(n, mask)
    nothing
  end
end

end
